@@ -493,6 +493,11 @@ sub chebyshev_psi {
   _validate_positive_integer($n);
   return Math::Prime::Util::PP::chebyshev_psi($n);
 }
+sub ramanujan_tau {
+  my($n) = @_;
+  _validate_positive_integer($n);
+  return Math::Prime::Util::PP::ramanujan_tau($n);
+}
 
 sub is_power {
   my($n, $a, $refp) = @_;
@@ -606,13 +611,18 @@ sub forperm (&$;$) {    ## no critic qw(ProhibitSubroutinePrototypes)
 
 sub vecreduce (&@) {    ## no critic qw(ProhibitSubroutinePrototypes)
   my($sub, @v) = @_;
-  my $count = shift @v;
+
+  # Mastering Perl page 162, works with old Perl
+  my $caller = caller();
+  no strict 'refs'; ## no critic(strict)
+  local(*{$caller.'::a'}) = \my $a;
+  local(*{$caller.'::b'}) = \my $b;
+  $a = shift @v;
   for my $v (@v) {
-    no strict 'refs'; ## no critic(strict)
-    local( ${caller() . '::a'}, ${caller() . '::b'} ) = ($count, $v);
-    $count = $sub->();
+    $b = $v;
+    $a = $sub->();
   }
-  $count;
+  $a;
 }
 
 sub vecextract {
@@ -622,6 +632,10 @@ sub vecextract {
   return Math::Prime::Util::PP::vecextract(@_);
 }
 
+sub sumdigits {
+  return Math::Prime::Util::PP::sumdigits(@_);
+}
+
 1;
 
 __END__
