@@ -4,7 +4,7 @@ use warnings;
 
 BEGIN {
   $ntheory::AUTHORITY = 'cpan:DANAJ';
-  $ntheory::VERSION = '0.52';
+  $ntheory::VERSION = '0.53';
 }
 
 BEGIN {
@@ -133,6 +133,7 @@ See L<Math::Prime::Util> for complete documentation.
   vecmax(@list)                       maximum of list of integers
   vecreduce { ... } @list             reduce / left fold applied to list
   vecextract(\@list, mask)            select from list based on mask
+  sumdigits(n)                        sum of digits, with optional base
   is_power(n)                         return k if n = p^k for integer p, max k
   is_power(n,k)                       return 1 if n = p^k for integer p and k
   is_power(n,k,\$root)                as above but set root to p.
@@ -164,6 +165,7 @@ See L<Math::Prime::Util> for complete documentation.
   znprimroot(n)                       smallest primitive root
   chebyshev_theta(n)                  first Chebyshev function
   chebyshev_psi(n)                    second Chebyshev function
+  ramanujan_tau(n)                    Ramanujan's Tau function
   consecutive_integer_lcm(n)          lcm(1 .. n)
   lucasu(P, Q, k)                     U_k for Lucas(P,Q)
   lucasv(P, Q, k)                     V_k for Lucas(P,Q)
