@@ -7,7 +7,7 @@
  *
  * Prime counts using the extended Lagarias-Miller-Odlyzko combinatorial method.
  *
- * Copyright (c) 2013 Dana Jacobsen (dana@acm.org)
+ * Copyright (c) 2013-2014 Dana Jacobsen (dana@acm.org)
  * This is free software; you can redistribute it and/or modify it under
  * the same terms as the Perl 5 programming language system itself.
  *
@@ -293,8 +293,8 @@ static UV _phi_recurse(UV x, UV a) {
   UV i, c = (a > PHIC) ? PHIC : a;
   UV sum = tablephi(x, c);
   if (a > c) {
-    UV p  = _XS_nth_prime(c);
-    UV pa = _XS_nth_prime(a);
+    UV p  = nth_prime(c);
+    UV pa = nth_prime(a);
     for (i = c+1; i <= a; i++) {
       UV xp;
       p = next_prime(p);
@@ -336,20 +336,42 @@ static IV _phi(UV x, UV a, int sign, const uint32_t* const primes, const uint32_
     cache[a*PHICACHEX+x] = sign * sum;
   return sum;
 }
-UV legendre_phi(UV x, UV a) {
-  /* TODO: tune these */
-  if ( (x > PHIC && a > 200) || (x > 1000000000 && a > 30) ) {
+UV legendre_phi(UV x, UV a)
+{
+  /* If 'x' is very small, give a quick answer with any 'a' */
+  if (x <= PHIC)
+    return tablephi(x, (a > PHIC) ? PHIC : a);
+
+  /* Shortcuts for large values, from R. Andrew Ohana */
+  if (a > (x >> 1))  return 1;
+  /* If a > prime_count(2^32), then we need not be concerned with composite
+   * x values with all factors > 2^32, as x is limited to 64-bit. */
+  if (a > 203280221) {  /* prime_count(2**32) */
+    UV pc = _XS_LMO_pi(x);
+    return (a > pc)  ?  1  :  pc - a + 1;
+  }
+  /* If a is large enough, check the ratios */
+  if (a > 1000000 && x < a*21) {  /* x always less than 2^32 */
+    if ( _XS_LMO_pi(x) < a)  return 1;
+  }
+
+  /* TODO:  R. Andrew Ohana's 2011 SAGE code is faster as the a value
+   * increases.  It uses a primelist as in the caching code below, as
+   * well as a binary search prime count on it (like in our lehmer). */
+
+  if ( a > 254 || (x > 1000000000 && a > 30) ) {
     uint16_t* cache;
     uint32_t* primes;
     uint32_t lastidx;
     UV res, max_cache_a = (a >= PHICACHEA) ? PHICACHEA : a+1;
     Newz(0, cache, PHICACHEX * max_cache_a, uint16_t);
-    primes = make_primelist(_XS_nth_prime(a+1), &lastidx);
+    primes = make_primelist(nth_prime(a+1), &lastidx);
     res = (UV) _phi(x, a, 1, primes, lastidx, cache);
     Safefree(primes);
     Safefree(cache);
     return res;
   }
+
   return _phi_recurse(x, a);
 }
 /****************************************************************************/
