@@ -35,8 +35,21 @@ static const unsigned char distancewheel30[30] =
 static const unsigned char wheeladvance30[30] =
     {1,6,5,4,3,2,1,4,3,2,1,2,1,4,3,2,1,2,1,4,3,2,1,6,5,4,3,2,1,2};
 /* subtract this from n to get to the previous wheel location */
-static const unsigned char wheelretreat[30] =
+static const unsigned char wheelretreat30[30] =
     {1,2,1,2,3,4,5,6,1,2,3,4,1,2,1,2,3,4,1,2,1,2,3,4,1,2,3,4,5,6};
+/* Given a sieve byte, this indicates the first zero */
+static const unsigned char nextzero30[256] =
+  {0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,
+   0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,6,0,1,0,2,0,1,0,3,0,1,0,2,
+   0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,
+   0,2,0,1,0,3,0,1,0,2,0,1,0,7,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,
+   0,1,0,2,0,1,0,5,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,
+   0,6,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,5,0,1,0,2,
+   0,1,0,3,0,1,0,2,0,1,0,4,0,1,0,2,0,1,0,3,0,1,0,2,0,1,0,8};
+/* At this m (p-30*(p/30)), OR with this to clear previous entries */
+static const unsigned char clearprev30[30] =
+    {  0,  0,  1,  1,  1,  1,  1,  1,  3,  3,  3,  3,  7,  7, 15,
+      15, 15, 15, 31, 31, 63, 63, 63, 63,127,127,127,127,127,127};
 
 
 #ifdef FUNC_is_prime_in_sieve
@@ -52,19 +65,20 @@ static int is_prime_in_sieve(const unsigned char* sieve, UV p) {
 /* Will return 0 if it goes past lastp */
 static UV next_prime_in_sieve(const unsigned char* sieve, UV p, UV lastp) {
   UV d, m;
+  unsigned char s;
   if (p < 7)
     return (p < 2) ? 2 : (p < 3) ? 3 : (p < 5) ? 5 : 7;
+  p++;
+  if (p >= lastp) return 0;
   d = p/30;
   m = p - d*30;
-  do {
-    if (m != 29) {
-      m = nextwheel30[m];
-    } else {
-      d++; m = 1;
-      if (d*30 >= lastp) return 0; /* sieves have whole bytes filled */
-    }
-  } while (sieve[d] & masktab30[m]);
-  return d*30+m;
+  s = sieve[d] | clearprev30[m];
+  while (s == 0xFF) {
+    d++;
+    if (d*30 >= lastp) return 0;
+    s = sieve[d];
+  }
+  return d*30 + wheel30[nextzero30[s]];
 }
 #endif
 #ifdef FUNC_prev_prime_in_sieve
@@ -88,28 +102,28 @@ static UV prev_prime_in_sieve(const unsigned char* sieve, UV p) {
     const unsigned char* sieve_ = sieve; \
     UV base_ = base; \
     UV p = a-base_; \
-    UV l_ = b-base_; \
+    UV l_ = b; \
     UV d_ = p/30; \
-    UV lastd_ = l_/30; \
-    UV mask_ = masktab30[ p-d_*30 + distancewheel30[ p-d_*30 ] ]; \
-    UV s_ = sieve_[d_]; \
-    while (d_ <= lastd_ && (s_ & mask_)) { \
-      mask_ <<= 1;  if (mask_ > 128) { s_ = sieve_[++d_]; mask_ = 1; } \
-    } \
-    p = d_*30 + imask30[mask_]; \
-    while ( p <= l_ ) { \
-      p += base_;
+    UV lastd_ = (l_-base_)/30; \
+    unsigned char bit_, s_ = sieve_[d_] | clearprev30[p-d_*30]; \
+    base_ += d_*30; \
+    while (1) { \
+      if (s_ == 0xFF) { \
+        do { \
+          base_ += 30;  d_++; \
+          if (d_ > lastd_) break; \
+          s_ = sieve_[d_]; \
+        } while (s_ == 0xFF); \
+        if (d_ > lastd_) break; \
+      } \
+      bit_ = nextzero30[s_]; \
+      s_ |= 1 << bit_; \
+      p = base_ + wheel30[bit_]; \
+      if (p > l_ || p < base_) break; \
+      { \
 
 #define END_DO_FOR_EACH_SIEVE_PRIME \
-      do { \
-        mask_ <<= 1; \
-        if (mask_ > 128) { \
-          while (++d_ <= lastd_) { s_ = sieve_[d_]; if (s_ != 0xFF) break; } \
-          if (d_ > lastd_) break; \
-          mask_ = 1; \
-        } \
-      } while (s_ & mask_); \
-      p = d_*30 + imask30[mask_]; \
+      } \
     } \
   }
 
@@ -119,37 +133,35 @@ static UV prev_prime_in_sieve(const unsigned char* sieve, UV p) {
     UV p  = a; \
     UV l_ = b; \
     UV d_ = p/30; \
-    UV s_, mask_ = 2; \
     UV lastd_ = l_/30; \
+    unsigned char s_, bit_; \
     get_prime_cache(l_, &sieve_); \
-    s_ = sieve_[d_]; \
-    if (p <= 5) { \
-      p = (p <= 2) ? 2 : (p <= 3) ? 3 : 5; \
-    } else if (p != 7) { \
-      mask_ = masktab30[ p-d_*30 + distancewheel30[ p-d_*30 ] ]; \
-      while (d_ <= lastd_ && (s_ & mask_)) { \
-        mask_ <<= 1;  if (mask_ > 128) { s_ = sieve_[++d_]; mask_ = 1; } \
+    if (p == 2) p = 1; \
+    s_ = sieve_[d_] | clearprev30[p-d_*30]; \
+    while (1) { \
+      if (p < 5) { \
+        p = (p < 2) ? 2 : (p < 3) ? 3 : 5; \
+      } else { \
+        if (s_ == 0xFF) { \
+          do { \
+            d_++; \
+            if (d_ > lastd_) break; \
+            s_ = sieve_[d_]; \
+          } while (s_ == 0xFF); \
+          if (d_ > lastd_) break; \
+        } \
+        bit_ = nextzero30[s_]; \
+        s_ |= 1 << bit_; \
+        p = d_*30 + wheel30[bit_]; \
+        if (p < d_*30) break; \
       } \
-      p = d_*30 + imask30[mask_]; \
-    } \
-    while ( p <= l_ ) {
+      if (p > l_) break; \
+      { \
 
 #define RETURN_FROM_EACH_PRIME(retstmt) \
-    do { release_prime_cache(sieve_); retstmt; } while (0)
+        do { release_prime_cache(sieve_); retstmt; } while (0)
 
 #define END_DO_FOR_EACH_PRIME \
-      if (p < 7) { \
-        p += 1 + (p > 2); \
-      } else { \
-        do { \
-          mask_ <<= 1; \
-          if (mask_ > 128) { \
-            if (++d_ > lastd_) break; \
-            s_ = sieve_[d_]; \
-            mask_ = 1; \
-          } \
-        } while (s_ & mask_); \
-        p = d_*30 + imask30[mask_]; \
       } \
     } \
     release_prime_cache(sieve_); \
