@@ -70,14 +70,20 @@ extern UV carmichael_lambda(UV n);
 extern UV jordan_totient(UV k, UV n);
 extern UV znprimroot(UV n);
 extern UV znorder(UV a, UV n);
+extern int is_primitive_root(UV a, UV p);
+#define is_square_free(n)  (moebius(n) != 0)
 
 extern IV stirling2(UV n, UV m);
 extern IV stirling1(UV n, UV m);
 
+extern IV ramanujan_tau(UV n);
+
 extern int strnum_minmax(int min, char* a, STRLEN alen, char* b, STRLEN blen);
 
 extern int is_catalan_pseudoprime(UV n);
 
+extern UV gcdz(UV x, UV y);
+
 #if defined(FUNC_isqrt) || defined(FUNC_is_perfect_square)
 static UV isqrt(UV n) {
   UV root;
@@ -116,6 +122,10 @@ static UV icbrt(UV n) {
 #endif
 
 #if defined(FUNC_gcd_ui) || defined(FUNC_lcm_ui)
+/* If we have a very fast ctz, then use the fast FLINT version of gcd */
+#if defined(__GNUC__) && (__GNUC__ >= 4 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
+#define gcd_ui(x,y) gcdz(x,y)
+#else
 static UV gcd_ui(UV x, UV y) {
   UV t;
   if (y < x) { t = x; x = y; y = t; }
@@ -125,6 +135,7 @@ static UV gcd_ui(UV x, UV y) {
   return x;
 }
 #endif
+#endif
 
 #ifdef FUNC_lcm_ui
 static UV lcm_ui(UV x, UV y) {
